@@ -45,7 +45,9 @@
 #include <freetds/checks.h>
 #include <freetds/bytes.h>
 #include <freetds/alloca.h>
-#include "replacements.h"
+#include <freetds/encodings.h>
+#include <freetds/enum_cap.h>
+#include <freetds/replacements.h>
 
 /** \cond HIDDEN_SYMBOLS */
 #define USE_ICONV (tds->conn->use_iconv)
@@ -93,7 +95,7 @@ static TDSRET tds_process_end(TDSSOCKET * tds, int marker, /*@out@*/ int *flags_
 static TDSRET tds_get_data_info(TDSSOCKET * tds, TDSCOLUMN * curcol, int is_param);
 static /*@observer@*/ const char *tds_token_name(unsigned char marker);
 static void adjust_character_column_size(TDSSOCKET * tds, TDSCOLUMN * curcol);
-int determine_adjusted_size(const TDSICONV * char_conv, int size);
+static int determine_adjusted_size(const TDSICONV * char_conv, int size);
 static /*@observer@*/ const char *tds_pr_op(int op);
 static int tds_alloc_get_string(TDSSOCKET * tds, /*@special@*/ char **string, size_t len) /*allocates *string*/;
 
@@ -252,7 +254,7 @@ tds_process_default_tokens(TDSSOCKET * tds, int marker)
 		tok_size = tds_get_byte(tds);
 		if (tok_size >= 3) {
 			tds_get_byte(tds);
-			tds5_negotiate_set_msg_type(tds, tds->conn->authentication, tds_get_usmallint(tds));
+			tds5_negotiate_set_msg_type(tds->conn->authentication, tds_get_usmallint(tds));
 			tok_size -= 3;
 		}
 		tds_get_n(tds, NULL, tok_size);
@@ -386,7 +388,7 @@ tds_process_loginack(TDSSOCKET *tds, TDSRET *login_succeeded)
 	 * TDS 4.2 reports 1 on success and is not
 	 * present on failure
 	 */
-	if (ack == 5 || ack == 1) {
+	if (ack == 5 || ack == 1 || (IS_TDS50(tds->conn) && ack == 0x85)) {
 		*login_succeeded = TDS_SUCCESS;
 		/* authentication is now useless */
 		if (tds->conn->authentication) {
@@ -837,7 +839,10 @@ tds_process_tokens(TDSSOCKET *tds, TDS_INT *result_type, int *done_flags, unsign
 
 	set_return_exit:
 		if (TDS_FAILED(rc)) {
-			tds_set_state(tds, rc == TDS_CANCELLED ? TDS_PENDING : TDS_DEAD);
+			if (rc == TDS_CANCELLED)
+				tds_set_state(tds, TDS_PENDING);
+			else
+				tds_close_socket(tds);
 			return rc;
 		}
 
@@ -2195,7 +2200,6 @@ tds_process_env_chg(TDSSOCKET * tds)
 	char *newval = NULL;
 	char **dest;
 	int new_block_size;
-	int lcid;
 	int memrc = 0;
 
 	CHECK_TDS_EXTRA(tds);
@@ -2231,8 +2235,7 @@ tds_process_env_chg(TDSSOCKET * tds)
 		} else {
 			tds_get_n(tds, tds->conn->collation, 5);
 			tds_get_n(tds, NULL, size - 5);
-			lcid = TDS_GET_UA4LE(tds->conn->collation) & 0xffffflu;
-			tds7_srv_charset_changed(tds->conn, tds->conn->collation[4], lcid);
+			tds7_srv_charset_changed(tds->conn, tds->conn->collation);
 		}
 		tdsdump_dump_buf(TDS_DBG_NETWORK, "tds->conn->collation now", tds->conn->collation, 5);
 		/* discard old one */
@@ -2742,30 +2745,6 @@ tds5_process_dyn_result2(TDSSOCKET * tds)
 	return tds_alloc_row(info);
 }
 
-/**
- * tds_get_token_size() returns the size of a fixed length token
- * used by tds_process_cancel() to determine how to read past a token
- * \param marker token type.
- */
-int
-tds_get_token_size(int marker)
-{
-	/* TODO finish */
-	switch (marker) {
-	case TDS_DONE_TOKEN:
-	case TDS_DONEPROC_TOKEN:
-	case TDS_DONEINPROC_TOKEN:
-		return 8;
-	case TDS_RETURNSTATUS_TOKEN:
-		return 4;
-	case TDS_PROCID_TOKEN:
-		return 8;
-	default:
-		return 0;
-	}
-}
-
-
 /**
  * tds_process_compute_names() processes compute result sets.
  * \tds
@@ -3111,6 +3090,7 @@ tds_prtype(int type)
 		TYPE(SYBTIME, "time");
 		TYPE(SYB5BIGTIME, "bigtime");
 		TYPE(SYB5BIGDATETIME, "bigdatetime");
+		TYPE(SYBMSTABLE, "user-defined table type");
 	default:
 		break;
 	}
@@ -3225,20 +3205,33 @@ adjust_character_column_size(TDSSOCKET * tds, TDSCOLUMN * curcol)
 	CHECK_TDS_EXTRA(tds);
 	CHECK_COLUMN_EXTRA(curcol);
 
-	if (is_unicode_type(curcol->on_server.column_type))
-		curcol->char_conv = tds->conn->char_convs[client2ucs2];
+	if (is_ascii_type(curcol->on_server.column_type)) {
+		/* don't override setting from column collation */
+		if (!curcol->char_conv)
+			curcol->char_conv = tds->conn->char_convs[client2server_chardata];
+		goto compute;
+	}
+
+	if (IS_TDS7_PLUS(tds->conn)) {
+		if (is_unicode_type(curcol->on_server.column_type))
+			curcol->char_conv = tds->conn->char_convs[client2ucs2];
+		goto compute;
+	}
 
 	/* Sybase UNI(VAR)CHAR fields are transmitted via SYBLONGBINARY and in UTF-16 */
-	if (curcol->on_server.column_type == SYBLONGBINARY && (
-		curcol->column_usertype == USER_UNICHAR_TYPE ||
-		curcol->column_usertype == USER_UNIVARCHAR_TYPE)) {
-#ifdef WORDS_BIGENDIAN
-		static const char sybase_utf[] = "UTF-16BE";
-#else
-		static const char sybase_utf[] = "UTF-16LE";
-#endif
+	if (is_unicode_type(curcol->on_server.column_type) ||
+		(curcol->on_server.column_type == SYBLONGBINARY && (
+		 curcol->column_usertype == USER_UNICHAR_TYPE ||
+		 curcol->column_usertype == USER_UNIVARCHAR_TYPE))) {
+		const int canonic_client = tds->conn->char_convs[client2ucs2]->from.charset.canonic;
+		const int sybase_utf16 = TDS_CHARSET_UTF_16LE;
 
-		curcol->char_conv = tds_iconv_get(tds->conn, tds->conn->char_convs[client2ucs2]->from.charset.name, sybase_utf);
+		if (tds_capability_has_res(tds->conn, TDS_RES_IMAGE_NONCHAR)) {
+			curcol->char_conv = tds_iconv_get_info(tds->conn, canonic_client, TDS_CHARSET_UTF_8);
+			goto compute;
+		}
+
+		curcol->char_conv = tds_iconv_get_info(tds->conn, canonic_client, sybase_utf16);
 
 		/* fallback to UCS-2LE */
 		/* FIXME should be useless. Does not works always */
@@ -3246,9 +3239,7 @@ adjust_character_column_size(TDSSOCKET * tds, TDSCOLUMN * curcol)
 			curcol->char_conv = tds->conn->char_convs[client2ucs2];
 	}
 
-	if (!curcol->char_conv && is_ascii_type(curcol->on_server.column_type))
-		curcol->char_conv = tds->conn->char_convs[client2server_chardata];
-
+compute:
 	if (!USE_ICONV || !curcol->char_conv)
 		return;
 
@@ -3276,12 +3267,17 @@ adjust_character_column_size(TDSSOCKET * tds, TDSCOLUMN * curcol)
  * \param size unconverted byte size
  * \return maximum size for converted string
  */
-int
+static int
 determine_adjusted_size(const TDSICONV * char_conv, int size)
 {
 	if (!char_conv)
 		return size;
 
+	/* same charset */
+	if ((char_conv->flags & TDS_ENCODING_MEMCPY) != 0
+	    || char_conv->to.charset.canonic == char_conv->from.charset.canonic)
+		return size;
+
 	/* avoid possible overflow */
 	if (size >= 0x10000000)
 		return 0x7fffffff;
